#include "command_buffers.h"
#include "gpu.h"
#include "transfers.h"
#include "resources.h"

#include "gfx/renderer_api.h"

#include "runtime/config.h"
#include "runtime/threading.h"

#include <stdlib.h>
#include <string.h>

RT_CVAR_I(rt_VkMaxBufferCount, "Number of slots for gpu buffers. Default: 1024.", 1024);

typedef struct rt_buffer_data_s {
    rt_buffer data;
    uint32_t version;
    struct rt_buffer_data_s *next_free;
} rt_buffer_data;

static rt_buffer_data *_buffers;
static rt_buffer_data *_first_free;
static rt_mutex *_list_lock;

rt_result InitBufferManagement(void) {
    size_t n = (size_t)rt_VkMaxBufferCount.i;
    _buffers = calloc(n, sizeof(rt_buffer_data));
    if (!_buffers)
        return RT_OUT_OF_MEMORY;
    _first_free = &_buffers[1];
    for (size_t i = 1; i < n - 1; ++i)
        _buffers[i].next_free = &_buffers[i + 1];
    _list_lock = rtCreateMutex();
    return RT_SUCCESS;
}

void ShutdownBufferManagement(void) {
    for (int i = 0; i < rt_VkMaxBufferCount.i; ++i) {
        if (_buffers[i].data.buffer == VK_NULL_HANDLE)
            continue;
        vmaDestroyBuffer(g_gpu.allocator, _buffers[i].data.buffer, _buffers[i].data.allocation);
        rtDestroyRWLock(&_buffers[i].data.lock);
        memset(&_buffers[i], 0, sizeof(_buffers[i]));
    }
    free(_buffers);
    _first_free = NULL;
    rtDestroyMutex(_list_lock);
}

void RT_RENDERER_API_FN(DestroyBuffers)(uint32_t count, rt_buffer_handle *buffers);

static void UploadViaMap(rt_buffer_data *buffer, const void *data, size_t size) {
    rtLockWrite(&buffer->data.lock);
    void *dev_mem = NULL;
    if (vmaMapMemory(g_gpu.allocator, buffer->data.allocation, &dev_mem) != VK_SUCCESS) {
        rtReportError("vk", "Unable to map buffer for upload");
        rtUnlockWrite(&buffer->data.lock);
        return;
    }
    memcpy(dev_mem, data, size);
    vmaUnmapMemory(g_gpu.allocator, buffer->data.allocation);
    if (!buffer->data.coherent)
        vmaFlushAllocation(g_gpu.allocator, buffer->data.allocation, 0, VK_WHOLE_SIZE);
    rtUnlockWrite(&buffer->data.lock);
}

/* Convenience function that decides between mapping or uploading via transfer buffer */
static void UploadData(rt_buffer_data *buffer, const void *data, size_t size) {
    if (buffer->data.mappable)
        UploadViaMap(buffer, data, size);
    else
        rtUploadToBuffer(buffer->data.buffer,
                         buffer->data.allocation,
                         buffer->data.owner,
                         data,
                         size);
}

rt_result RT_RENDERER_API_FN(CreateBuffers)(uint32_t count,
                                            const rt_buffer_info *info,
                                            rt_buffer_handle *p_buffers) {
    for (uint32_t i = 0; i < count; ++i) {
        rtLockMutex(_list_lock);
        rt_buffer_data *slot = _first_free;
        if (!slot) {
            rtUnlockMutex(_list_lock);
            if (i > 0)
                rtRenDestroyBuffers(i, p_buffers);
            return RT_OUT_OF_MEMORY;
        }
        _first_free = slot->next_free;
        rtUnlockMutex(_list_lock);

        VkBufferUsageFlags buffer_usage = 0;
        switch (info->type) {
        case RT_BUFFER_TYPE_VERTEX:
            buffer_usage = VK_BUFFER_USAGE_VERTEX_BUFFER_BIT;
            break;
        case RT_BUFFER_TYPE_INDEX:
            buffer_usage = VK_BUFFER_USAGE_INDEX_BUFFER_BIT;
            break;
        case RT_BUFFER_TYPE_STORAGE:
            buffer_usage = VK_BUFFER_USAGE_STORAGE_BUFFER_BIT;
            break;
        case RT_BUFFER_TYPE_UNIFORM:
            buffer_usage = VK_BUFFER_USAGE_UNIFORM_BUFFER_BIT;
            break;
        }
        buffer_usage |= VK_BUFFER_USAGE_TRANSFER_DST_BIT;

        VkBufferCreateInfo buffer_info = {
            .sType       = VK_STRUCTURE_TYPE_BUFFER_CREATE_INFO,
            .size        = info->size,
            .usage       = buffer_usage,
            .sharingMode = VK_SHARING_MODE_EXCLUSIVE,
        };

        VmaMemoryUsage alloc_usage           = 0;
        VmaAllocationCreateFlags alloc_flags = 0;
        switch (info->usage) {
        case RT_BUFFER_USAGE_STATIC:
            alloc_usage = VMA_MEMORY_USAGE_AUTO_PREFER_DEVICE;
            alloc_flags = 0;
            break;
        case RT_BUFFER_USAGE_DYNAMIC:
            alloc_usage = VMA_MEMORY_USAGE_AUTO;
            alloc_flags = VMA_ALLOCATION_CREATE_HOST_ACCESS_SEQUENTIAL_WRITE_BIT |
                          VMA_ALLOCATION_CREATE_HOST_ACCESS_ALLOW_TRANSFER_INSTEAD_BIT;
            break;
        case RT_BUFFER_USAGE_TRANSIENT:
            alloc_usage = VMA_MEMORY_USAGE_AUTO_PREFER_DEVICE;
            alloc_flags = VMA_ALLOCATION_CREATE_HOST_ACCESS_SEQUENTIAL_WRITE_BIT;
            break;
        }
        VmaAllocationCreateInfo alloc_info = {.usage = alloc_usage, .flags = alloc_flags};

        VkResult res = vmaCreateBuffer(g_gpu.allocator,
                                       &buffer_info,
                                       &alloc_info,
                                       &slot->data.buffer,
                                       &slot->data.allocation,
                                       NULL);
        if (res != VK_SUCCESS) {
            rtReportError("vk", "Failed to create a buffer: %u", res);
            rtLockMutex(_list_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            rtUnlockMutex(_list_lock);
            if (i > 0)
                rtRenDestroyBuffers(i, p_buffers);
            return RT_UNKNOWN_ERROR;
        }

        rt_create_rwlock_result lock_res = rtCreateRWLock();
        if (!lock_res.ok) {
            rtReportError("vk", "Failed to create lock for buffer.");
            vmaDestroyBuffer(g_gpu.allocator, slot->data.buffer, slot->data.allocation);
            rtLockMutex(_list_lock);
            slot->next_free = _first_free;
            _first_free     = slot;
            rtUnlockMutex(_list_lock);
            if (i > 0)
                rtRenDestroyBuffers(i, p_buffers);
            return RT_UNKNOWN_ERROR;
        }

        VkMemoryPropertyFlags properties;
        vmaGetAllocationMemoryProperties(g_gpu.allocator, slot->data.allocation, &properties);
        slot->data.mappable = (properties & VK_MEMORY_PROPERTY_HOST_VISIBLE_BIT) != 0;
        slot->data.coherent = (properties & VK_MEMORY_PROPERTY_HOST_COHERENT_BIT) != 0;
        
        slot->data.owner = RT_VK_UNOWNED;
        slot->data.state = RT_BUFFER_STATE_NOT_USED;

        if (info->data)
            UploadData(slot, info->data, info->size);

        ptrdiff_t index      = slot - _buffers;
        p_buffers[i].index   = (uint32_t)index;
        p_buffers[i].version = slot->version;
    }
    return RT_SUCCESS;
}

void RT_RENDERER_API_FN(DestroyBuffers)(uint32_t count, rt_buffer_handle *buffers) {
    for (uint32_t i = 0; i < count; ++i) {
        if (buffers[i].index >= (uint32_t)rt_VkMaxBufferCount.i)
            continue;
        rt_buffer_data *slot = &_buffers[buffers[i].index];
        if (slot->version != buffers[i].version) {
            rtLog("vk", "Tried to destroy a buffer with an invalid handle (version mismatch).");
            continue;
        }
        rtLockWrite(&slot->data.lock);
        slot->version = (slot->version + 1) & RT_RENDER_BACKEND_HANDLE_MAX_VERSION;
        vmaDestroyBuffer(g_gpu.allocator, slot->data.buffer, slot->data.allocation);
        slot->data.buffer = VK_NULL_HANDLE;
        slot->data.allocation = VK_NULL_HANDLE;
        rtUnlockWrite(&slot->data.lock);
        rtDestroyRWLock(&slot->data.lock);

        rtLockMutex(_list_lock);
        slot->next_free = _first_free;
        _first_free     = slot;
        rtUnlockMutex(_list_lock);
    }
}

rt_buffer *rtGetBuffer(rt_buffer_handle handle) {
    if (handle.index >= (uint32_t)rt_VkMaxBufferCount.i)
        return NULL;
    rt_buffer_data *slot = &_buffers[handle.index];
    if (slot->version != handle.version) {
        rtLog("vk", "Tried to access a buffer with an invalid handle (version mismatch).");
        return NULL;
    }
    return &slot->data;
}